/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * Contributor(s): Eric Murphy   , ericmurphy@jabber.org
 *                 Michael Hearn , tweedledee@jabber.org
 *                 Jim Ray       , jray@lanscape.net
 *                 Jan Hecking   , jhecking@netgaroo.com
 * 2002.03.05
 *
 **********************************************************************************/

const CONTRACTID  = "@jabberxm.mozdev.org;1";
const CID         = Components.ID("{a4cdcfae-1dd1-11b2-9c25-9b5d19103742}");
const CiJabberXM  = Components.interfaces.JabberXM;
const nsISupports = Components.interfaces.nsISupports;

/**********************************************************************************/
// This initializes the subcomponents, and these mostly handle incoming data.

function initialize() {
  if (JabberXMModule.socket == null) {
    jabberXMSax =
      new Components.Constructor("@jabberxm.mozdev.org/sax;1","JabberXMSax");

    JabberXMModule.stream = new jabberXMSax();
    JabberXMModule.stream.setDocumentHandler(new Handler());

    jabberXMSocket =
      new Components.Constructor("@jabberxm.mozdev.org/socket;1","JabberXMSocket");

    JabberXMModule.socket = new jabberXMSocket();
    JabberXMModule.socket.stream(JabberXMModule.stream);

    JabberXMModule.session = new Session();
    JabberXMModule.iq      = new IQ();
    JabberXMModule.message = new Message();
  }
}

/**********************************************************************************/
// HANDLER: This separates out the different packet types.

function Handler() {
  this.wrappedJSObject = this;

  this.packetHandler = function(packet) {
    if(packet.name == "message") {
      JabberXMModule.message.packetMessage(packet);
    }
    else if (packet.name == "iq") {
      JabberXMModule.iq.packetIQ(packet);
    }
    else if (packet.name == "presence") {
      JabberXMModule.presence.packetPresence(packet);
    }
    else if (packet.name == "stream:stream") {
      JabberXMModule.session.packetSession(packet);
    }
  }
}

/**********************************************************************************/
// MESSAGE: This class handles incoming <message> packets, and other functions.

function Message() {
  this.postHandler = null;
  // Array of namespaces each containing an array of handlers for that namespace.
  this.nsHandler = new Array(); 
}

Message.prototype.initChat = function(window,chatWindowUri,attributes) {
  this.chatWindow=window;
  this.chatWindowUri=chatWindowUri;
  this.chatAttributes=attributes;
}

Message.prototype.initMessage = function(window,messageWindowUri,attributes) {
  this.messageWindow=window;
  this.messageWindowUri=messageWindowUri;
  this.messageAttributes=attributes;
}

Message.prototype.setChatHandler = function(chatHandler) {
  this.chathandler = chatHandler.wrappedJSObject.chatHandler;
}

Message.prototype.setMessageHandler = function(messageHandler) {
  this.messagehandler = messageHandler.wrappedJSObject.messageHandler;
}

Message.prototype.handleChat = function(from,subject,body,html) {
  if(JabberXMModule.chatWindows[from] == null) {
    win = JabberXMModule.chatWindows[from] = 
          this.chatWindow.open(this.chatWindowUri,from,this.chatAttributes);

    JabberXMModule.chatWindows[from] = win;
    JabberXMModule.chatWindows[from].chatHandler = this.chatHandler;

    JabberXMModule.chatWindows[from].subject = subject;
    JabberXMModule.chatWindows[from].body = body;
    JabberXMModule.chatWindows[from].html = html;
  }
  else // PRE is Garbage
    JabberXMModule.chatWindows[from].PREchatHandler(from,subject,body,html);

  JabberXMModule.chatWindows[from].getAttention();
}

Message.prototype.handleMessage = function(from,subject,body)
{
}

Message.prototype.openChat = function(jid) {
  if(JabberXMModule.chatWindows[jid] == null) {
    JabberXMModule.chatWindows[jid] = 
      this.chatWindow.open(this.chatWindowUri,jid,this.chatAttributes);

    JabberXMModule.chatWindows[jid].chatHandler = this.chatHandler;
  }
}

Message.prototype.closeChat = function(jid) {
  JabberXMModule.chatWindows[jid] = null;
}

Message.prototype.openMessage = function(jid) {
  if(JabberXMModule.messageWindows[jid] == null) {
    JabberXMModule.messageWindows[jid] = 
      this.messageWindow.open(this.messageWindowUri,jid,this.messageAttributes);

    JabberXMModule.messageWindows[jid].messageHandler = this.messageHandler;
  }
}

Message.prototype.messageChat = function(jid) {
  JabberXMModule.chatWindows[jid] = null;
}

Message.prototype.sendChat = function(jid,subject,body,html,x) {
  if(subject != '' || subject != null)
    subject = '<subject>' + subject + '</subject>';

  if(x == null) x = '';
  JabberXMModule.socket.write (
    '<message to="' + jid + '" type="chat">' + 
      subject +
      '<body>' + body + '</body>' +
      '<html xmlns="http://www.w3.org/1999/xhtml">' + html + '</html>' +
      x + 
    '</message>'
  );
}

Message.prototype.sendGroupChat = function(jid,subject,body,html,thread,x) {
  if(subject != '' || subject != null)
    subject = '<subject>' + subject + '</subject>';

  JabberXMModule.socket.write (
    '<message to="' + jid + '" type="groupchat">' + 
      '<thread>' + thread + '</thread>' +
      subject +
      '<body>' + body + '</body>' +
      html +
    '</message>'
  );
}

Message.prototype.sendMessage = function(jid,subject,body,html,x) {
  if(subject != '')
    subject = '<subject>' + subject + '</subject>';

  if(x == null) x = '';
  JabberXMModule.socket.write (
    '<message to="' + jid + '">' + 
      subject +
      '<body>' + body + '</body>' +
      '<html>' + html + '</html>' +
      x +
    '</message>'
  );
}

Message.prototype.packetMessage = function(packet)

{

   var messageType = '';



   switch( packet.attribs["type"] )

   {

      case 'chat':

         messageType = 'chat';

         break;

      case 'groupchat':

         messageType = 'groupchat';

         break;

      case 'error':

         messageType = 'error';

         break;

      default:

         messageType = 'normal';

   }



   var subject=''; var body=''; var thread=''; var html=''; var w='';



   var calledHandler = false;



   for(var i = 0; i < packet.children.length; i++)

   {

      var child = packet.children[i];

      switch( child.name )

      {

         case "subject":

            subject = child.cdata;

            break;



         case "body":

            body = child.cdata;

            break;



         case "thread":

            thread = child.cdata;

            break;



         case "html":

            var currentData = JabberXMModule.socket.getData();

            var beginning = currentData.search("<html");

            html = currentData.substring(beginning);

            var end = html.search("</html>");

            html = html.substring(0,end+7);

            break;

         case "w":

            var currentData = JabberXMModule.socket.getData();

            var beginning = currentData.search("<w");

            w = currentData.substring(beginning);

            var end = w.search("</w>");

            w = w.substring(0,end+4);

            break;



         case "x":

            var xmlns = child.attribs["xmlns"];

            if( this.nsHandler[ xmlns ] )

            {

               var handler = this.nsHandler[ xmlns ];

               for( var i = 0; i < handler.length; i++ )

               {

                  handler[i].handleMessage( packet );

                  calledHandler = true;

               }

            }

            else

            {

               dump( "No handler registered for namespace " + xmlns + "." );

            }

            break;



         default:

            dump( "unknown message element: " + child.name );

      }

   }



   if( calledHandler )

      return;


   var fulljid =packet.attribs["from"];

   var jid = packet.attribs["from"] + "/";

   jid = jid.substr( 0, jid.indexOf( "/" ) );

   if(messageType=='normal' || messageType=='chat')

   {

        //this.handleMessage(jid,subject,body);

         this.handleChat(jid,subject,body,html);

   }

   else if(messageType=='groupchat')

   {
      if(w != '')
      {
         name = JabberXMModule.conferenceUserNameList[fulljid];
         this.whiteboardHandler.handleWhiteboard(name,body,w);
      }
      else if(JabberXMModule.conferenceChatHandlers[jid] != null)
      {
        name = JabberXMModule.conferenceUserNameList[fulljid];
        if(name==null)
          name='';

        JabberXMModule.conferenceChatHandlers[jid](name,subject,body,html);
      }
      else if(this.postHandler != null)

      {

         this.postHandler.handlePost(jid,subject,body,thread);

      }

   }

   else

   {



   }

}

Message.prototype.registerNSHandler = function( namespace, handler )

{

   var newHandler = handler.wrappedJSObject;

   if( this.nsHandler[ namespace ] )

   {

      if( this.nsHandler[ namespace ][ newHandler ] )

      {

	 dump( "Handler already registered." );

	 return false;

      }

      else

      {

	 this.nsHandler[ namespace ].push( newHandler );

      }

   }

   else

   {

      this.nsHandler[ namespace ] = new Array( newHandler );

   }

   return true;

}



Message.prototype.unregisterNSHandler = function( namespace, handler )

{

  var oldHandler = handler.wrappedJSObject;

  if( this.nsHandler[ namespace ] )

  {

     for( var i = 0; i < this.nsHandler[ namespace ].length; i++ )

     {

	if( this.nsHandler[ namespace ][ i ] == oldHandler )

	{

	   this.nsHandler[ namespace ].splice( i, 1 );

           if( this.nsHandler[ namespace ].length == 0 )

           {

             //TODO: delete the namespace entry - something like this:

             //this.nsHandler[ namespace ] = undefined;

           }

	   return true;

	}

     }

  }

  return false;

}

/**********************************************************************************/
// IQ: This class handles <iq> packets coming in.

function IQ(){}

IQ.prototype.packetIQ = function(packet)
{
  if( (packet.attribs['type'] == 'result') && (packet.attribs['id'] == 'register') )
  {
    JabberXMModule.session.postRegistration();
  }
  else if( (packet.attribs['type'] == 'error') && (packet.attribs['id'] == 'register') )
  {
    JabberXMModule.session.registrationError();
  }
  else
  {
    for(var i = 0; i < packet.children.length; i++)
    {
      if (packet.children[i].name=="vCard")
      {
         vCard = packet.children[i];

         // Works, with lag
         for (var x=0; x < vCard.children.length; x++)
         {
            if (vCard.children[x].name=='EMAIL')
            {
               var Email =  (vCard.children[x].cdata);

//             window._content.location = ( 'mailto:' + Email + '?Subject=File Transfer' );

            }

         }
      }

      else if(packet.children[i].name == "query")
      {
         if (packet.children[i].attribs["xmlns"]=="jabber:iq:roster")
         {
            JabberXMModule.roster.packetRoster(packet.children[i]);
         }
         else if (packet.children[i].attribs["xmlns"]=="jabber:iq:agents")
         {
//            this.agents.packetAgents(packet.children[i]);
            dump('Agents listing received.');
         }
         else if (packet.children[i].attribs["xmlns"]=="jabber:iq:agent")
         {
            dump('Agent properties received.');
         }
         else if (packet.children[i].attribs["xmlns"]=="jabber:iq:oob")
         {
            oob = packet.children[i];
            dump('Out of band received.');

            // Need to handle this when detached
//            window._content.location = packet.children[0].children[0].cdata;
//            writeIQ(packet.attribs["from"],"result");

            // Need to close file download
         }
         else if (packet.children[i].attribs["xmlns"]=="jabber:iq:time")
         {
            time = packet.children[i];
            dump('Time received.');
         }
         else if (packet.children[i].attribs["xmlns"]=="jabber:iq:version")
         {
            version = packet.children[i];
            dump('Version received.');
         }
         else
         { 
            dump('Unknown query received.');
         }
      }
      else if(packet.children[i].name == "user" || packet.children[i].name == "conference")
      {
        jid = packet.attribs["from"];

        if(JabberXMModule.conferenceRosters[jid] == null) {
          JabberXMModule.conferenceRosters[jid] = new Roster();
        }
        JabberXMModule.conferenceRosters[jid].packetRoster(packet);
      }
   }

  }
}

/**********************************************************************************/
// PRESENCE: This class handles <presence> packets coming in and going out.

function Presence()
{
   this.presence = new DataSequence('http://jabberxm.mozdev.org/rdf#groups');
   this.showResource   = 'http://jabberxm.mozdev.org/rdf#show';
   this.statusResource = 'http://jabberxm.mozdev.org/rdf#status';
   this.presenceLoaded = false;
}

Presence.prototype.sendPresence = function(show, status)
{
   if (show == 'available')
   {
     JabberXMModule.socket.write('<presence/>');
   }
   else
   {
      JabberXMModule.socket.write
      (
         '<presence><show>' + 
           show   + '</show><status>' + 
           status + '</status></presence>'
      );
   }
}

Presence.prototype.changePresence = function(jid,res,show,status)
{
   this.presenceLoaded = true;

   if(res!='')
     var user = this.presence.rdf.GetResource(jid+'/'+res);
   else
     var user = this.presence.rdf.GetResource(jid);

   target = 
      this.presence.datasource.GetTarget
      (
         user,
         this.presence.rdf.GetResource(this.showResource),
         true
      );

   if ( target == null )
   {
      this.presence.datasource.Assert
      (
         user,
         this.presence.rdf.GetResource(this.showResource),
         this.presence.rdf.GetLiteral(show),
         true
      );
   }
   else if ( JabberXMModule.hasReceivedRoster )
   {
      this.presence.datasource.Change
      (
         user,
         this.presence.rdf.GetResource(this.showResource),
         target,
         this.presence.rdf.GetLiteral(show)
      );
   }

   target = 
      this.presence.datasource.GetTarget
      (
         user,
         this.presence.rdf.GetResource(this.statusResource),
         true
      );

   if ( target == null )
   {
      this.presence.datasource.Assert
      (
         user,
         this.presence.rdf.GetResource(this.statusResource),
         this.presence.rdf.GetLiteral(status),
         true
      );
   }
   else if ( JabberXMModule.hasReceivedRoster )
   {
      this.presence.datasource.Change
      (
         user,
         this.presence.rdf.GetResource(this.statusResource),
         target,
         this.presence.rdf.GetLiteral(status)
      );
   }

   this.presence.sequence.AppendElement(user);

   if(res!='') {
     res='';
     this.changePresence(jid,res,show,status);
   }
}

/*
   this.presence.datasource.Unassert
   (
      user,
      this.presence.rdf.GetResource(this.showResource),
      this.presence.rdf.GetLiteral(show)
   );
*/



Presence.prototype.packetPresence = function(packet)
{
         // Get the JID and Resource (hack)
         p = '';  q = 0;

   if(packet.attribs["type"]=='error')
   {}
   else if(packet.attribs["type"]=='subscribe')
   {
      jid = packet.attribs["from"]; status = '';

      for(var i = 0; i < packet.children.length; i++){
         if(packet.children[i].name == "status")
            status = packet.children[i].cdata;
      }
      this.presenceRequestWindow.open
      (this.presenceRequestWindowUri,jid,this.presenceRequestAttributes);
   }

   else if (packet.attribs["from"] != null)
   {
      from = packet.attribs["from"];

      // Hack off resource, except for Agents

      if (from.search('icq') != -1 && from.search('@icq') == -1)
      {
         jid = from; res = '';
      }
      else if (from.search('aim') != -1 && from.search('@aim') == -1)
      {
         jid = from; res = '';
      }
      else if (from.search('msn') != -1 && from.search('@msn') == -1)
      {
         jid = from; res = '';
      }
      else if (from.search('yahoo') != -1 && from.search('@yahoo') == -1)
      {
         jid = from; res = '';
      }
      else
      {
         from = packet.attribs["from"] + '/';

         while (p != '/')
         {
            p = from[q++];
         }

         jid = (from.slice(0,q-1));

         res = (from.slice(q,from.length-1));
      }
   }

   // Get the Priority
   for(var i = 0; i < packet.children.length; i++)
   {
      if(packet.children[i].name == "priority")
      {
         priority = packet.children[i].cdata;
      }
   }

   // Default
   show = 'available';

   if (packet.attribs["type"] == 'error')
   {
      show = 'unknown';
   }

   if (packet.attribs["type"] == 'unavailable')
   {
      show = 'unavailable';
   }

   for(var i = 0; i < packet.children.length; i++)
   {
      if(packet.children[i].name == "show")
      {
         if (packet.children[i].cdata != '')
            show = packet.children[i].cdata;
      }
   }

   if(show == 'available')
     status = 'Online';
   else
     status = '';

   for(var i = 0; i < packet.children.length; i++)
   {
      if(packet.children[i].name == "status")
      {
         status = packet.children[i].cdata;
      }
   }

   this.changePresence(jid,res,show,status);
}


/**********************************************************************************/
// SESSION: 

function Session()
{
   this.id = ''; // session id for SHA-1 digest
   this.isAuth = false;
}

Session.prototype.packetSession = function(packet)
{
   if(!this.isAuth)
   {
      this.id = packet.attribs["id"];
      this.authenticate();
      this.getRoster();
   }
   else
   {
      this.id = packet.attribs["id"];
      this.register();
   }
}

Session.prototype.authenticate = function()
{
   var digest = this.id + this.password;
   digest = calcSHA1(digest);

   JabberXMModule.socket.write
   (
      "<iq type='set' id='auth'><query xmlns='jabber:iq:auth'>" + 
      "<username>" + this.username + "</username>" + 
      "<digest>" + digest + "</digest>" + 
      "<priority>2</priority>" +
      "<resource>" + this.resource + "</resource>" + 
      "</query></iq>" + "<presence/>"
   );
   JabberXMModule.isConnected = true;
}

Session.prototype.postRegistration = function()
{
   try
   {
      this.postRegister.handler(this.username);
   }
   catch(e){dump(e);}
}

Session.prototype.registrationError = function()
{
   try
   {
      this.registerError.handler(this.username);
   }
   catch(e){dump(e);}
}

Session.prototype.getRoster = function()
{
   JabberXMModule.socket.write
   (
      "<iq id='doroster' type='get'><query xmlns='jabber:iq:roster'/></iq>"
   );
}

Session.prototype.register = function()
{
   JabberXMModule.socket.write
   (
     '<iq type="set" id="register"><query xmlns="jabber:iq:register">'
     + '<username>' + this.username + '</username>' 
     + '<password>' + this.password + '</password>'
     + '</query></iq>'
   );
}

/**********************************************************************************/
// This created the hash used for the password on the login of the client.

var hex_chr = "0123456789abcdef";

function hex(num) {
   var str = "";
   for(var j = 7; j >= 0; j--)
      str += hex_chr.charAt((num >> (j * 4)) & 0x0F);
   return str;
}

function str2blks_SHA1(str) {
  var nblk = ((str.length + 8) >> 6) + 1;
  var blks = new Array(nblk * 16);
  for(var i = 0; i < nblk * 16; i++) blks[i] = 0;
  for(i = 0; i < str.length; i++)
    blks[i >> 2] |= str.charCodeAt(i) << (24 - (i % 4) * 8);
  blks[i >> 2] |= 0x80 << (24 - (i % 4) * 8);
  blks[nblk * 16 - 1] = str.length * 8;
  return blks;
}

function add(x, y) {
  return ((x&0x7FFFFFFF) + (y&0x7FFFFFFF)) ^ (x&0x80000000) ^ (y&0x80000000);
}

function rol(num, cnt) {
  return (num << cnt) | (num >>> (32 - cnt));
}

function ft(t, b, c, d) {
  if(t < 20) return (b & c) | ((~b) & d);
  if(t < 40) return b ^ c ^ d;
  if(t < 60) return (b & c) | (b & d) | (c & d);
  return b ^ c ^ d;
}

function kt(t) {
  if(t < 20) return 0x5A827999;
  if(t < 40) return 0x6ED9EBA1;
  if(t < 60) return 0x8F1BBCDC;
  return 0xCA62C1D6;
}

function calcSHA1(str) {
  var x = str2blks_SHA1(str);
  var w = new Array(80);

  var a = 0x67452301;
  var b = 0xEFCDAB89;
  var c = 0x98BADCFE;
  var d = 0x10325476;
  var e = 0xC3D2E1F0;

  for(var i = 0; i < x.length; i += 16) {
    var olda = a;
    var oldb = b;
    var oldc = c;
    var oldd = d;
    var olde = e;

    for(var j = 0; j < 80; j++) {
      if(j < 16) w[j] = x[i + j];
      else w[j] = rol(w[j-3] ^ w[j-8] ^ w[j-14] ^ w[j-16], 1);
      t = add(add(rol(a, 5), ft(j, b, c, d)), add(add(e, w[j]), kt(j)));
      e = d;
      d = c;
      c = rol(b, 30);
      b = a;
      a = t;
    }
    a = add(a, olda);
    b = add(b, oldb);
    c = add(c, oldc);
    d = add(d, oldd);
    e = add(e, olde);
  }
  return hex(a) + hex(b) + hex(c) + hex(d) + hex(e);
}

/**********************************************************************************/
// ROSTER: 

function Roster()
{
   this.rosteritemsResourceURI = 'http://jabberxm.mozdev.org/rdf#rosteritems';
   this.labelResourceURI       = 'http://jabberxm.mozdev.org/rdf#label';
   this.typeResourceURI        = 'http://jabberxm.mozdev.org/rdf#type';

   this.roster = new DataSequence(this.rosteritemsResourceURI);
   this.groups = new Object(); // associative array
}

Roster.prototype.setUser = function(jid,nickname,show,status,group)
{
   var contactResource = this.roster.rdf.GetResource(jid);

   this.roster.datasource.Assert
   (
      contactResource,
      this.roster.rdf.GetResource(this.labelResourceURI),
      this.roster.rdf.GetLiteral(nickname),
      true
   );

   this.roster.datasource.Assert
   (
      contactResource,
      this.roster.rdf.GetResource(this.typeResourceURI),
      this.roster.rdf.GetLiteral('contact'),
      true
   );

   this.groups[group].AppendElement(contactResource);
   JabberXMModule.presence.changePresence(jid,'',show,status);
}

Roster.prototype.removeUser = function(jid,group)
{
   var contactResource = this.roster.rdf.GetResource(jid);

   this.groups[group].RemoveElement(contactResource,true);
}

Roster.prototype.setGroup = function(group)
{
   if (this.groups[group] == null)
   {
                                                       // Moz bug 105759
      var groupResource = this.roster.rdf.GetResource('chrome://jabberzilla/content/main.xul#' + group);

      this.roster.datasource.Assert
      (
         groupResource,
         this.roster.rdf.GetResource(this.labelResourceURI),
         this.roster.rdf.GetLiteral(group),
         true
      );

      this.roster.datasource.Assert
      (
         groupResource,
         this.roster.rdf.GetResource(this.labelResourceURI),
         this.roster.rdf.GetLiteral(group),
         true
      );

      this.roster.datasource.Assert
      (
         groupResource,
         this.roster.rdf.GetResource(this.typeResourceURI),
         this.roster.rdf.GetLiteral('group'),
         true
      );

      contactsAnonymousResource = this.roster.rdf.GetAnonymousResource();

      this.roster.datasource.Assert
      (
         groupResource,
         this.roster.rdf.GetResource(this.rosteritemsResourceURI),
         contactsAnonymousResource,
         true
      );

      this.groups[group] = 
         this.roster.rdfc.MakeSeq(this.roster.datasource,contactsAnonymousResource);

      this.roster.sequence.AppendElement(groupResource);
   }
}

Roster.prototype.removeGroup = function(group)
{
   if (this.groups[group] != null)
   {
       //Remove Group Resource
   }
}

// frozen - eric@mozdev.org - 2001.10.16
Roster.prototype.changeTransportNames = function(user)
{
   if(user.jid.search('icq') != -1 && user.jid.search('@icq') == -1) {
      user.nickname = ('ICQ Network');
   }
   else if (user.jid.search('aim') != -1 && user.jid.search('@aim') == -1) {
      user.nickname = ('AIM Network');
   }
   else if (user.jid.search('msn') != -1 && user.jid.search('@msn') == -1) {
      user.nickname = ('MSN Network');
   }
   else if (user.jid.search('yahoo') != -1 && user.jid.search('@yahoo') == -1) {
      user.nickname = ('Yahoo Network');
   }
   return user;
}

Roster.prototype.packetRoster = function(packet)
{
  JabberXMModule.hasReceivedRoster = true;
  var user = new Object();
  for(var h = 0; h < packet.children.length; h++) {
    if(packet.children[h].name == "item") {
      user.jid = packet.children[h].attribs["jid"];
      user.jid = user.jid.toLowerCase();
      user.nickname = packet.children[h].attribs["name"];

      user = this.changeTransportNames(user);
      child = packet.children[h];

      for(var j = 0; j < child.children.length; j++) {
        if(child.children[j].name == 'group') {
          user.group = child.children[j].cdata;
        }
      }
      if(user.group == null) user.group = 'Default';
      this.setGroup(user.group);

      subscription = packet.children[h].attribs["subscription"];

      if(subscription == 'to' || subscription == 'both') {
        this.setUser(user.jid,user.nickname,'unavailable','',user.group);
      }
      else {
        this.setUser(user.jid,user.nickname,'unknown','unknown',user.group);
      }
    }
    // Handle current user's nickname
    else if(packet.children[h].name == "user") {

      user.jid = packet.children[h].attribs["jid"];
      user.nickname = packet.children[h].attribs["name"];
      user.group = "Group";

      if(packet.children[h].attribs["type"] == "remove")
      {
        this.removeUser(user.jid,user.group);
      }
      else
      {
        this.setGroup(user.group);
        this.setUser(user.jid,user.nickname,'available','Online',user.group);
        JabberXMModule.conferenceUserNameList[user.jid] = user.nickname;
      }
    }
    // Handle conference roster users
    else if(packet.children[h].name == "conference") {
      for(var j = 0; j < packet.children[h].children.length; j++) {
        users = packet.children[h].children;

        user.jid = users[j].attribs["jid"];
        user.nickname = users[j].attribs["name"];
        user.group = "Group";

        this.setGroup(user.group);
        this.setUser(user.jid,user.nickname,'available','Online',user.group);
        JabberXMModule.conferenceUserNameList[user.jid] = user.nickname;
      }
    }
    user.group = null; // Reset group
  }
}

/**********************************************************************************/
// This is an object for creating a list of data in RDF.

function DataSequence(uri) {
  this.rdf = 
    Components.classes['@mozilla.org/rdf/rdf-service;1'].
    getService(Components.interfaces.nsIRDFService);

  this.datasource = 
    Components.classes['@mozilla.org/rdf/datasource;1?name=in-memory-datasource'].
    createInstance(Components.interfaces.nsIRDFDataSource);  

  this.root = this.rdf.GetResource('urn:root');
  this.anon = this.rdf.GetAnonymousResource();

  this.datasource.Assert(this.root,this.rdf.GetResource(uri),this.anon,true);

  this.rdfc = 
    Components.classes['@mozilla.org/rdf/container-utils;1'].
    getService(Components.interfaces.nsIRDFContainerUtils);

  this.sequence = this.rdfc.MakeSeq(this.datasource,this.anon);

  this.rdfs = 
    Components.classes['@mozilla.org/rdf/container;1'].
    getService(Components.interfaces.nsIRDFContainer);
}

/**********************************************************************************/
// XML

JabberXM.prototype.sendXML = function(xml)
{
   if (xml != null || xml != '')
   {
      JabberXMModule.socket.write(xml);
      return true;
   }
   else
      return false;
}

JabberXM.prototype.setViewStream = function(viewStream)
{
   viewStream = viewStream.wrappedJSObject;
   JabberXMModule.socket.setViewStream(viewStream);
}

/*************************************/
// SESSION

JabberXM.prototype.getJID = function()
{
   return (JabberXMModule.username + '@' + JabberXMModule.server);
}

JabberXM.prototype.setRegister = function(username,password)
{
   JabberXMModule.session.username = username;
   JabberXMModule.session.password = password;
   JabberXMModule.session.isAuth = true;
}

JabberXM.prototype.setPostRegisterHandler = function(handler)
{
   // Will be called after sucessful registration
   JabberXMModule.session.postRegister = handler.wrappedJSObject;
}

JabberXM.prototype.setRegisterErrorHandler = function(handler)
{
   // Will be called after sucessful registration
   JabberXMModule.session.registerError = handler.wrappedJSObject;
}

JabberXM.prototype.setAuthenticate = function(username,password,resource,priority)
{
   if(resource == "" || resource == null)
   { this.resource = ("jabberzilla"); }

   if(priority == "" || priority == null)
   { this.priority = 0; }

   JabberXMModule.session.username = username;
   JabberXMModule.session.password = password;
   JabberXMModule.session.resource = resource;
   JabberXMModule.session.priority = priority;
   JabberXMModule.session.isAuth   = false;
}

JabberXM.prototype.authenticate = function()
{
   // Used after registration
   JabberXMModule.session.authenticate();
}

JabberXM.prototype.connect = function(server)
{
   if (JabberXMModule.isConnected)
   {
      return false;
   }
   else
   {
      JabberXMModule.socket.open(server,5222);
      JabberXMModule.socket.initAsyncRead();

      JabberXMModule.socket.write
      (
        "<stream:stream to='" + server
        + "' xmlns:stream='http://etherx.jabber.org/streams'"
        + "  xmlns='jabber:client'>"
      );
      JabberXMModule.server = server;
      return true;
   }
}

JabberXM.prototype.isConnected = function()
{
   return JabberXMModule.isConnected;
}

JabberXM.prototype.disconnect = function()
{
   if(JabberXMModule.isConnected)
      JabberXMModule.socket.write("<presence type='unavailable'/></stream:stream>");

   JabberXMModule.socket.close();
   JabberXMModule.isConnected = false;
   JabberXMModule.hasReceivedRoster = false;

   JabberXMModule.socket   = null;
   JabberXMModule.stream   = null;
   JabberXMModule.roster   = null;
   JabberXMModule.presence = null;
   JabberXMModule.server   = '';
   JabberXMModule.username = '';

   JabberXMModule.roster = new Roster('unknown');
   JabberXMModule.presence = new Presence();

   JabberXMModule.chatWindows = new Object();
   JabberXMModule.messageWindows = new Object();

   initialize(); // reset everything
}

JabberXM.prototype.sessionTimeLength = function()
{
}

JabberXM.prototype.changePassword = function(oldpass,newpass)
{
}

/*************************************/
// ROSTER

JabberXM.prototype.hasReceivedRoster = function()
{
   return JabberXMModule.hasReceivedRoster;
}

JabberXM.prototype.requestRoster = function()
{
}

JabberXM.prototype.getRosterDataSource = function() {
   return JabberXMModule.roster.roster.datasource;
}

JabberXM.prototype.addRosterGroup = function(group) {
   JabberXMModule.roster.setGroup(group);
}

JabberXM.prototype.removeRosterGroup = function(group) {
   JabberXMModule.roster.removeGroup(group);
}

/*************************************/
// PRESENCE

JabberXM.prototype.setPresence = function(show,status)
{
   JabberXMModule.presence.sendPresence(show,status);
}

JabberXM.prototype.sendPresence = function(jid)
{
   JabberXMModule.socket.write('<presence to="' + jid + '"/>');
}

JabberXM.prototype.getPresenceDataSource = function()
{
   return JabberXMModule.presence.presence.datasource;
}

JabberXM.prototype.subscribePresence = function(jid)
{
   JabberXMModule.socket.write('<presence to="'+jid+'" type="subscribe"/>');
}

JabberXM.prototype.subscribedPresence = function(jid)
{
   JabberXMModule.socket.write('<presence to="'+jid+'" type="subscribed"/>');
}

JabberXM.prototype.unsubscribePresence = function(jid)
{
   JabberXMModule.socket.write('<presence to="'+jid+'" type="unsubscribe"/>');
}

JabberXM.prototype.unsubscribedPresence = function(jid)
{
   JabberXMModule.socket.write('<presence to="'+jid+'" type="unsubscribed"/>');
}

JabberXM.prototype.initPresenceRequest = function(window,uri,attributes)
{
   JabberXMModule.presence.presenceRequestWindow=window;
   JabberXMModule.presence.presenceRequestWindowUri=uri;
   JabberXMModule.presence.presenceRequestAttributes=attributes;
}

/*************************************/
// CONTACT

JabberXM.prototype.addContact = function(jid,nickname,group,reason)
{
  JabberXMModule.socket.write
  (
    '<iq id="addContact" type="set">' 
    + '<query xmlns="jabber:iq:roster">'
    + '<item jid="'+jid+'" name="'+nickname+'"><group>'+group+'</group></item>'
    + '</query>'
    + '</iq>'
  );

  JabberXMModule.socket.write
  (
    '<presence to="'+jid+'" type="subscribe"><status>'+reason+'</status></presence>'
  );
}

JabberXM.prototype.removeContact = function(jid,group) // UPDATE nick too?
{
  JabberXMModule.socket.write
  (
    '<iq id="delContact" type="set">' 
    + '<query xmlns="jabber:iq:roster">'
    + '<item ask="subscribe" jid="'+jid+'" name="'+nickname+'" subscription="remove"><group>'+group+'</group></item>'
    + '</query>'
    + '</iq>'
  );
}

JabberXM.prototype.setNickname = function(jid,nickname)
{
}

/*************************************/
// MESSAGE

JabberXM.prototype.setChatHandler = function(chatHandler)
{
   JabberXMModule.message.setChatHandler(chatHandler);
}

JabberXM.prototype.setMessageHandler = function(messageHandler)
{
   JabberXMModule.message.setMessageHandler(messageHandler);
}

JabberXM.prototype.openChat = function(jid)
{
   JabberXMModule.message.openChat(jid);
}

JabberXM.prototype.openMessage = function(jid)
{
   JabberXMModule.message.openMessage(jid);
}

JabberXM.prototype.closeChat = function(jid)
{
   JabberXMModule.message.closeChat(jid);
}

JabberXM.prototype.closeMessage = function(jid)
{
   JabberXMModule.message.closeMessage(jid);
}

JabberXM.prototype.sendChat = function(jid,subject,body,html,x)
{
   JabberXMModule.message.sendChat(jid,subject,body,html,x);
}


JabberXM.prototype.sendGroupChat = function(jid,subject,body,html,thread,x)
{
   JabberXMModule.message.sendGroupChat(jid,subject,body,html,thread,x);
}

JabberXM.prototype.sendMessage = function(jid,subject,body,html,x)
{
   JabberXMModule.message.sendMessage(jid,subject,body,html,x);
}

JabberXM.prototype.initChat = function(window,uri,attributes)
{
   JabberXMModule.message.initChat(window,uri,attributes);
}

JabberXM.prototype.initMessage = function(window,uri,attributes)
{
   JabberXMModule.message.initMessage(window,uri,attributes);
}

JabberXM.prototype.initMessageSound = function(soundUri)
{
}

/*************************************/
// NAMESPACE

JabberXM.prototype.registerNSHandler = function( namespace, handler )

{

   JabberXMModule.message.registerNSHandler( namespace, handler );

}



JabberXM.prototype.unregisterNSHandler = function( namespace, handler )

{

   JabberXMModule.message.unregisterNSHandler( namespace, handler );

}

/*************************************/
// IQ

JabberXM.prototype.sendFile = function(url)
{
}

JabberXM.prototype.initDoneFile = function(doneWindowUri,attributes)
{
}

JabberXM.prototype.initReceiveFile = function(recieveWindowUri,attributes)
{
}

JabberXM.prototype.search = function(first,last,nickname,email)
{
}

JabberXM.prototype.getSearchDataSource = function()
{
}

JabberXM.prototype.getVCard = function(jid)
{
  JabberXMModule.socket.write('<iq type="get" to="' + jid + 
                              '"><VCARD xmlns="vcard-temp"/></iq>');
}

JabberXM.prototype.initClientVersion = function(version)
{
   initialize(); // HACK!
}

JabberXM.prototype.sendIQ = function(jid,type,attributes,contents) {

  iq="<iq to='" + jid + "' " + attributes + " type='" + type  + "'>" +contents+ "</iq>";
  JabberXMModule.socket.write(iq);
}

/*************************************/
// WEB CONTENT

JabberXM.prototype.setPostHandler = function(postHandler) {
  postHandler = postHandler.wrappedJSObject;
  JabberXMModule.message.postHandler = postHandler;
}

/*************************************/
// CONFERENCE

JabberXM.prototype.getConferenceRoster = function(room)
{
  return JabberXMModule.conferenceRosters[room].roster.datasource;
}

JabberXM.prototype.setConferenceChatHandler = function(room,chatHandler)
{
  JabberXMModule.conferenceChatHandlers[room] = chatHandler.wrappedJSObject.chatHandler;
}

/*************************************/
// WHITEBOARD

JabberXM.prototype.setWhiteboardHandler = function(whiteboardHandler)
{
   whiteboardHandler = whiteboardHandler.wrappedJSObject;

   JabberXMModule.message.whiteboardHandler = whiteboardHandler;
}

/*************************************/

JabberXM.prototype.QueryInterface = function(iid)
{
  if (!iid.equals(CiJabberXM))
    throw Components.results.NS_ERROR_NO_INTERFACE;
  return this;
}

function JabberXM(){};

/**********************************************************************************/

JabberXMFactory = new Object();

JabberXMFactory.createInstance = function(outer,iid)
{
  if (outer != null)
    throw Components.results.NS_ERROR_NO_AGGREGATION;

  if (!iid.equals(CiJabberXM) && !iid.equals(nsISupports))
    throw Components.results.NS_ERROR_INVALID_ARG;

  return new JabberXM();
}

/**********************************************************************************/

var JabberXMModule =
{
   isConnected       : false,
   hasReceivedRoster : false,

   username          : '',
   server            : '',

   socket            : null,
   stream            : null,

   message           : null,

   session           : null,

   iq                : null,
   presence          : new Presence(),
   roster            : new Roster(),

   conferenceRosters      : new Array(),
   conferenceChatHandlers : new Array(),
   conferenceUserNameList : new Array(),

   chatWindows       : new Object(),
   messageWindows    : new Object(),
}

JabberXMModule.registerSelf = function(compMgr,fileSpec,location,type)
{
  dump("*** Registering JabberXM.\n");
    
  compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  compMgr.registerFactoryLocation(CID,"JabberXM",CONTRACTID,fileSpec,location,type);
}

JabberXMModule.unregisterSelf = function(compMgr, fileSpec, location)
{
  compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  compMgr.unregisterFactoryLocation(CID,fileSpec);
}

JabberXMModule.getClassObject = function(compMgr,cid,iid)
{
  if(cid.equals(CID)) return JabberXMFactory;
    
  if(!iid.equals(Components.interfaces.nsIFactory))
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

  throw Components.results.NS_ERROR_NO_INTERFACE;
}

JabberXMModule.canUnload = function(compMgr)
{
  return true;
}

function NSGetModule(compMgr, fileSpec)
{
  return JabberXMModule;
}